// Copyright (c) 2021 Julien Floret
// Copyright (c) 2021 Robin Jarry
// SPDX-License-Identifier: BSD-3-Clause

$primary: #3762de;
$secondary: #ff8552;
$white: #fff;
$lightgray: #e5e7eb;
$darkgray: #374151;
$released: #edffeb;
$locked: #fffee8;
$padding: 0.5rem;
$width: 1400px;
$banner_height: 5rem;
$font_mono:
  ui-monospace,
  SFMono-Regular,
  Menlo,
  Monaco,
  Consolas,
  "Liberation Mono",
  "Courier New",
  monospace;
$font_sans:
  system-ui,
  -apple-system,
  'Segoe UI',
  Roboto,
  Helvetica,
  Arial,
  sans-serif,
  'Apple Color Emoji',
  'Segoe UI Emoji';
