{# Copyright (c) 2021 Julien Floret #}
{# Copyright (c) 2021 Robin Jarry #}
{# SPDX-License-Identifier: BSD-3-Clause #}
{% extends "base.html" %}

{% block breadcrumbs %}
{{breadcrumbs(['branches'])}}
{% endblock %}

{% block page_content %}
{% set width = branches|map(attribute='name')|map('length')|max|default(80) %}
{% set width = [width, 10]|max %}
<section class="branches" style="column-width: {{width * 0.8}}em;">
  {% for branch in branches|sort(attribute='name') %}
  <div>
    <a class="branch" href="{{branch.name}}/">{{branch.name}}</a>
  </div>
  {% endfor %}
</section>
{% endblock %}
